##################### generated by xml-casa (v2) from msuvbinflag.xml ###############
##################### 98ef26c0660306926ac8c8cbc6ae416b ##############################
from __future__ import absolute_import
from casashell.private.stack_manip import find_local as __sf__
from casashell.private.stack_manip import find_frame as _find_frame
from casatools.typecheck import validator as _pc
from casatools.coercetype import coerce as _coerce
from casatools.errors import create_error_string
from casatasks.private.task_msuvbinflag import msuvbinflag as _msuvbinflag_t
from casatasks.private.task_logging import start_log as _start_log
from casatasks.private.task_logging import end_log as _end_log
from collections import OrderedDict
import numpy
import sys
import os

import shutil

def static_var(varname, value):
    def decorate(func):
        setattr(func, varname, value)
        return func
    return decorate

class _msuvbinflag:
    """
    msuvbinflag ---- An algorithm to identify outliers in the UV plane through defined uniform grids by msuvbin.
    Both msuvbin and msuvbinflag tasks are experimental
    

    
    Msuvbinflag is an automatic flagging algorithm for the identification of Radio Frequency Interference (RFI)
    in the UV plane. The visibilities in input CASA Measurements Set (MS) has to be gridded into an uniformed UV plane
    with the msuvbin task.
    The principle undelying this task is that any interference that has effects on a large scale in the image domain will appear
    as a compact structure in UV-domain. No astronomical source will have that behavior especially not centred around (U=0, V=0).
    
    Multiple algorithms will be provided to locate these UV compact structures.
    
    

    --------- parameter descriptions ---------------------------------------------

    binnedvis    Name of input uniform binned visibility data (MS) set on UV plane using msuvbin task.
                 default: none;
                 example: binnedvis='SNR_Binnedvis.ms'
    method       There are four algorithm options proposed for selecting the outliers as flags (radial_per_plane, radial_mean_annular,  regionalMean, gradient)
                  
                  method='radial_per_plane': This algorithm works independently
                  on each plane of the binned ms. A mean radial profile is
                  gereated by doing a weighted average of all data at a given
                  UV-radius. A polynomial fit is done of this radial averaged
                  profile. All the uv points which have amplitude above the this
                  fitted profile at the radius of the uv-point by nsigma is
                  marked as flagged (originalValue[u,v] >
                  radiafitValue[r]+nsigma*rms).
                 
                 method='radial_mean_annular': The di Then average values are
                 calculated for each annular circle with one unit equal spacing.
                 A binary linear least square fitting is generated followed by
                 the comparison analysis between the original value and the
                 fitting value. A point is treated as an outliers to be flagged
                 if original value is greater than the sum of the fitting value
                 and the multiple of the user defined sigma and the interpolated
                 value.  eg: originalValue[x,y] >
                 radiafitValue[r]+sigma*interpolateValue[r].
                 
                 method='regionalMean': This algorithm is to compute the
                 statistical mean, median and standard deviation for every point
                 as a center within a square region which the size is defined by
                 user. A point is treated as an outliers to be flagged if it is
                 beyond the certain criteria threshold.  eg: originalValue[x,y])
                 > regionalMean+sigma*regionalStd
                 
                 method='gradient': This algorithm is to compute the gradient for
                 every point.  Then a point is treated as an outliers to be
                 flagged if there is a sudden change of gradient sign over a few
                 pixel in x and y directions.
                                
                 default: 'radial_per_plane'
                 example: method='radial_mean_annular'
    nsigma       User defined sigma value for comparison for radia algorithm. It is double type.
                 The default value is 5.
    sizeRegion   User defined local square regional size for "regionalMean" and "gradient" algorithm option.
    gradientflip User defined radius value for gradient algorithm. It is double type.
                 The default value is 1.
    doplot       Generate debugging plots; can slow down a lot.
                 The default value is False.
    [1;42mRETURNS[1;m         void

    --------- examples -----------------------------------------------------------

    
    
    
    msuvbinflag(binnedvis='SNR_G55_10s.ms', method='radial_per_plane', nsigma=5)
    
    Keyword arguments:
    binnedvis -- Name of input binned visibility file which is output from msuvbin.
    After msuvbinflag, binnedvis is modified with flagging info.
    
    method -- Three method options: "radian", "regional mean","median","gradient"
    default: 'radian'
    
    nsigma -- User set nsigma value for radial algorithm.
    default=5.0
    
    


    """

    _info_group_ = """flagging, manipulation, imaging"""
    _info_desc_ = """An algorithm to identify outliers in the UV plane through defined uniform grids by msuvbin.
      Both msuvbin and msuvbinflag tasks are experimental
  """

    __schema = {'binnedvis': {'type': 'cStr', 'coerce': _coerce.to_str}, 'method': {'type': 'cVariant', 'coerce': [_coerce.to_variant] # <allowed> IS NOT ALLOWED FOR A PARAMETER OF TYPE any
}, 'nsigma': {'type': 'cFloat', 'coerce': _coerce.to_float}, 'sizeRegion': {'type': 'cInt'}, 'gradientflip': {'type': 'cInt'}, 'doplot': {'type': 'cBool'}}

    def __init__(self):
        self.__stdout = None
        self.__stderr = None
        self.__root_frame_ = None

    def __globals_(self):
        if self.__root_frame_ is None:
            self.__root_frame_ = _find_frame( )
            assert self.__root_frame_ is not None, "could not find CASAshell global frame"
        return self.__root_frame_

    def __to_string_(self,value):
        if type(value) is str:
            return "'%s'" % value
        else:
            return str(value)

    def __validate_(self,doc,schema):
        return _pc.validate(doc,schema)

    def __do_inp_output(self,param_prefix,description_str,formatting_chars):
        out = self.__stdout or sys.stdout
        description = description_str.split( )
        prefix_width = 23 + 15 + 4
        output = [ ]
        addon = ''
        first_addon = True
        if len(description) == 0:
            out.write(param_prefix + " #\n")
            return
        while len(description) > 0:
            ## starting a new line.....................................................................
            if len(output) == 0:
                ## for first line add parameter information............................................
                if len(param_prefix)-formatting_chars > prefix_width - 1:
                    output.append(param_prefix)
                    continue
                addon = param_prefix + ' #'
                first_addon = True
                addon_formatting = formatting_chars
            else:
                ## for subsequent lines space over prefix width........................................
                addon = (' ' * prefix_width) + '#'
                first_addon = False
                addon_formatting = 0
            ## if first word of description puts us over the screen width, bail........................
            if len(addon + description[0]) - addon_formatting + 1 > self.term_width:
                ## if we're doing the first line make sure it's output.................................
                if first_addon: output.append(addon)
                break
            while len(description) > 0:
                ## if the next description word puts us over break for the next line...................
                if len(addon + description[0]) - addon_formatting + 1 > self.term_width: break
                addon = addon + ' ' + description[0]
                description.pop(0)
            output.append(addon)
        out.write('\n'.join(output) + '\n')

    #--------- return nonsubparam values ----------------------------------------------

    def __method_dflt( self, glb ):
        return 'radial_per_plane'

    def __method( self, glb ):
        if 'method' in glb: return glb['method']
        return 'radial_per_plane'

    def __doplot_dflt( self, glb ):
        return False

    def __doplot( self, glb ):
        if 'doplot' in glb: return glb['doplot']
        return False

    def __binnedvis_dflt( self, glb ):
        return ''

    def __binnedvis( self, glb ):
        if 'binnedvis' in glb: return glb['binnedvis']
        return ''



    #--------- return inp/go default --------------------------------------------------
    def __nsigma_dflt( self, glb ):
        if self.__method( glb ) == "radial_per_plane": return float(5)
        if self.__method( glb ) == "radial_mean_annular": return float(5)
        return None
    def __sizeRegion_dflt( self, glb ):
        if self.__method( glb ) == "regionalMean": return int(20)
        return None
    def __gradientflip_dflt( self, glb ):
        if self.__method( glb ) == "gradient": return int(1)
        return None

    #--------- return subparam values -------------------------------------------------
    def __nsigma( self, glb ):
        if 'nsigma' in glb: return glb['nsigma']
        dflt = self.__nsigma_dflt( glb )
        if dflt is not None: return dflt
        return float(5.0)
    def __sizeRegion( self, glb ):
        if 'sizeRegion' in glb: return glb['sizeRegion']
        dflt = self.__sizeRegion_dflt( glb )
        if dflt is not None: return dflt
        return int(20)
    def __gradientflip( self, glb ):
        if 'gradientflip' in glb: return glb['gradientflip']
        dflt = self.__gradientflip_dflt( glb )
        if dflt is not None: return dflt
        return int(1)

    #--------- subparam inp output ----------------------------------------------------
    def __binnedvis_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__binnedvis_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return ''
        description = 'Name of input binned MS from msuvbin task.'
        value = self.__binnedvis( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'binnedvis': value},{'binnedvis': self.__schema['binnedvis']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-15.15s = %s%-23s%s' % ('binnedvis',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __method_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__method_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return 'radial_per_plane'
        description = 'Flagging method option.(radial_per_plane, radial_mean_annular, gradient'
        value = self.__method( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'method': value},{'method': self.__schema['method']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('\x1B[1m\x1B[47m%-15.15s =\x1B[0m %s%-23s%s' % ('method',pre,self.__to_string_(value),post),description,13+len(pre)+len(post))
    def __nsigma_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__nsigma_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return float(5.0)
        if self.__nsigma_dflt( self.__globals_( ) ) is not None:
             description = 'User defined number of sigma values for comparison for radial algorithm and gradient algorithm.'
             value = self.__nsigma( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'nsigma': value},{'nsigma': self.__schema['nsigma']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('nsigma',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __sizeRegion_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__sizeRegion_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return int(20)
        if self.__sizeRegion_dflt( self.__globals_( ) ) is not None:
             description = 'User defined local square regional size.'
             value = self.__sizeRegion( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'sizeRegion': value},{'sizeRegion': self.__schema['sizeRegion']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('sizeRegion',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __gradientflip_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__gradientflip_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return int(1)
        if self.__gradientflip_dflt( self.__globals_( ) ) is not None:
             description = 'User defined number of pixels value for gradient flip.'
             value = self.__gradientflip( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'gradientflip': value},{'gradientflip': self.__schema['gradientflip']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('gradientflip',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __doplot_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__doplot_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return False
        description = 'if True will generate lots of plot for debugging purposes'
        value = self.__doplot( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'doplot': value},{'doplot': self.__schema['doplot']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-15.15s = %s%-23s%s' % ('doplot',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))

    #--------- global default implementation-------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def set_global_defaults(self):
        self.set_global_defaults.state['last'] = self
        glb = self.__globals_( )
        if 'method' in glb: del glb['method']
        if 'gradientflip' in glb: del glb['gradientflip']
        if 'sizeRegion' in glb: del glb['sizeRegion']
        if 'binnedvis' in glb: del glb['binnedvis']
        if 'nsigma' in glb: del glb['nsigma']
        if 'doplot' in glb: del glb['doplot']


    #--------- inp function -----------------------------------------------------------
    def inp(self):
        print("# msuvbinflag -- %s" % self._info_desc_)
        self.term_width, self.term_height = shutil.get_terminal_size(fallback=(80, 24))
        self.__binnedvis_inp( )
        self.__method_inp( )
        self.__nsigma_inp( )
        self.__sizeRegion_inp( )
        self.__gradientflip_inp( )
        self.__doplot_inp( )

    #--------- tget function ----------------------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def tget(self,savefile=None):
        from casashell.private.stack_manip import find_frame
        from runpy import run_path
        filename = savefile
        if filename is None:
            filename = "msuvbinflag.last" if os.path.isfile("msuvbinflag.last") else "msuvbinflag.saved"
        if os.path.isfile(filename):
            glob = find_frame( )
            newglob = run_path( filename, init_globals={ } )
            for i in newglob:
                glob[i] = newglob[i]
            self.tget.state['last'] = self
        else:
            print("could not find last file: %s\nsetting defaults instead..." % filename)
            self.set_global_defaults( )

    #--------- tput function ----------------------------------------------------------
    def tput(self,outfile=None):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s

        _postfile = outfile if outfile is not None else os.path.realpath('msuvbinflag.last')

        _invocation_parameters = OrderedDict( )
        _invocation_parameters['binnedvis'] = self.__binnedvis( self.__globals_( ) )
        _invocation_parameters['method'] = self.__method( self.__globals_( ) )
        _invocation_parameters['nsigma'] = self.__nsigma( self.__globals_( ) )
        _invocation_parameters['sizeRegion'] = self.__sizeRegion( self.__globals_( ) )
        _invocation_parameters['gradientflip'] = self.__gradientflip( self.__globals_( ) )
        _invocation_parameters['doplot'] = self.__doplot( self.__globals_( ) )

        try:
            with open(_postfile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-12s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#msuvbinflag( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: return False
        return True

    def __call__( self, binnedvis=None, method=None, nsigma=None, sizeRegion=None, gradientflip=None, doplot=None ):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s
        _prefile = os.path.realpath('msuvbinflag.pre')
        _postfile = os.path.realpath('msuvbinflag.last')
        task_result = None
        _arguments = [binnedvis,method,nsigma,sizeRegion,gradientflip,doplot]
        _invocation_parameters = OrderedDict( )
        if any(map(lambda x: x is not None,_arguments)):
            # invoke python style
            # set the non sub-parameters that are not None
            local_global = { }
            if binnedvis is not None: local_global['binnedvis'] = binnedvis
            if method is not None: local_global['method'] = method
            if doplot is not None: local_global['doplot'] = doplot

            # the invocation parameters for the non-subparameters can now be set - this picks up those defaults
            _invocation_parameters['binnedvis'] = self.__binnedvis( local_global )
            _invocation_parameters['method'] = self.__method( local_global )
            _invocation_parameters['doplot'] = self.__doplot( local_global )

            # the sub-parameters can then be set. Use the supplied value if not None, else the function, which gets the appropriate default
            _invocation_parameters['nsigma'] = self.__nsigma( _invocation_parameters ) if nsigma is None else nsigma
            _invocation_parameters['sizeRegion'] = self.__sizeRegion( _invocation_parameters ) if sizeRegion is None else sizeRegion
            _invocation_parameters['gradientflip'] = self.__gradientflip( _invocation_parameters ) if gradientflip is None else gradientflip

        else:
            # invoke with inp/go semantics
            _invocation_parameters['binnedvis'] = self.__binnedvis( self.__globals_( ) )
            _invocation_parameters['method'] = self.__method( self.__globals_( ) )
            _invocation_parameters['nsigma'] = self.__nsigma( self.__globals_( ) )
            _invocation_parameters['sizeRegion'] = self.__sizeRegion( self.__globals_( ) )
            _invocation_parameters['gradientflip'] = self.__gradientflip( self.__globals_( ) )
            _invocation_parameters['doplot'] = self.__doplot( self.__globals_( ) )
        try:
            with open(_prefile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-12s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#msuvbinflag( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: pass
        try:
            _logging_state_ = None
            assert _pc.validate(_invocation_parameters,self.__schema), create_error_string(_pc.errors)
            _logging_state_ = _start_log( 'msuvbinflag', [ 'binnedvis=' + repr(_pc.document['binnedvis']), 'method=' + repr(_pc.document['method']), 'nsigma=' + repr(_pc.document['nsigma']), 'sizeRegion=' + repr(_pc.document['sizeRegion']), 'gradientflip=' + repr(_pc.document['gradientflip']), 'doplot=' + repr(_pc.document['doplot']) ] )
            task_result = _msuvbinflag_t( _pc.document['binnedvis'],_pc.document['method'],_pc.document['nsigma'],_pc.document['sizeRegion'],_pc.document['gradientflip'],_pc.document['doplot'] )
        except Exception as e:
            from traceback import format_exc
            from casatasks import casalog
            casalog.origin('msuvbinflag')
            casalog.post("Exception Reported: Error in msuvbinflag: %s" % str(e),'SEVERE')
            casalog.post(format_exc( ))
            raise #exception is now raised
            #task_result = False
        finally:
            try:
                os.rename(_prefile,_postfile)
            except: pass
            if _logging_state_:
                task_result = _end_log( _logging_state_, 'msuvbinflag', task_result )

        #Added if _msuvbinflag_t returns False and does not raise an exception.
        if task_result is False:
            raise

        return task_result #Still needed

msuvbinflag = _msuvbinflag( )

